﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using Microsoft.SolverFoundation.Services;
using System.Data;
using System.Data.OleDb;
using System.Data.Linq;


namespace OMLFromCSharp
{
    /// <summary>
    /// Solve portfolio models
    /// </summary>
    [ClassInterface(ClassInterfaceType.AutoDual)]
    public class Portfolio
    {
        /// <summary>
        /// Version of the model
        /// </summary>
        string versionString = @"PORTFOLIO MODEL 1.0";

        /// <summary>
        /// Holds the QP model
        /// </summary>
        string strModelQP = @"Model[
              Parameters[Sets,I,T],
              Parameters[Reals,AdjRet[T,I],Mean[I]],
              Parameters[Reals,Lambda={0}],
              Parameters[Integers,NumT={1}],

              Decisions[Reals[0,Infinity],x[I]],
              Decisions[Reals,w[T],Return],

              Constraints[
                  Sum[{{i,I}},x[i]]==1,
                  Foreach[{{t,T}},w[t]==Sum[{{i,I}},AdjRet[t,i]*x[i]]],
                  Return==Sum[{{i,I}},Mean[i]*x[i]]
              ],

              Goals[Minimize[Obj->Lambda*Sum[{{t,T}},w[t]^2]/NumT-Return]]
           ]";

        /// <summary>
        /// Holds the LP model
        /// </summary>
        string strModelLP = @"Model[
              Parameters[Sets,I,T],
              Parameters[Reals,AdjRet[T,I],Mean[I]],
              Parameters[Reals,Lambda={0}],
              Parameters[Integers,NumT={1}],

              Decisions[Reals[0,Infinity],x[I],yplus[T],ymin[T]],
              Decisions[Reals,w[T],Return],

              Constraints[
                  Sum[{{i,I}},x[i]]==1,
                  Foreach[{{t,T}},w[t]==Sum[{{i,I}},AdjRet[t,i]*x[i]]],
                  Return==Sum[{{i,I}},Mean[i]*x[i]],
                  Foreach[{{t,T}},yplus[t]-ymin[t]==w[t]]
              ],

              Goals[Minimize[Obj->Lambda*Sum[{{t,T}},yplus[t]+ymin[t]]/NumT-Return]]
           ]";

        /// <summary>
        ///  SFS
        /// </summary>
        SolverContext context;

        /// <summary>
        /// msf solution
        /// </summary>
        Solution solution;

        /// <summary>
        /// Number of instruments we can invest in
        /// </summary>
        int numInstruments;
        
        /// <summary>
        /// number of time periods
        /// </summary>
        int numYears;

        /// <summary>
        /// lambda value used 
        /// </summary>
        double lambda;

        /// <summary>
        /// Connection string for MS Access 
        /// Use x86 architecture!
        /// </summary>
        string connection = @"Provider=Microsoft.ACE.OLEDB.12.0;Data Source=C:\projects\ms\OMLFromCSharp\OMLFromCSharp\bin\Debug\portfolio1.accdb;Persist Security Info=False;";

        /// <summary>
        /// Connection object
        /// </summary>
        OleDbConnection conn;

        /// <summary>
        /// list of instruments
        /// </summary>
        string[] instruments;

        /// <summary>
        ///  Constructor
        /// </summary>
        public Portfolio()
        {
            context = SolverContext.GetContext();
        }

        /// <summary>
        /// get query result as DataSet
        /// </summary>
        /// <param name="query">query as string</param>
        /// <returns></returns>
        private DataSet SelectOleDbSrvRows(string query)
        {
            DataSet ds = new DataSet();
            OleDbDataAdapter adapter = new OleDbDataAdapter();
            adapter.SelectCommand = new OleDbCommand(query, conn);
            adapter.Fill(ds);
            return ds;
        }

        /// <summary>
        /// Perform some magic to make sure the query output arrives in OML model.
        /// </summary>
        /// <param name="p">OML/SFS parameter</param>
        /// <param name="query">database query</param>
        /// <param name="valueColumn">column with values</param>
        /// <param name="IndexColumns">columns with indices</param>
        private void setBinding(Parameter p, string query, string valueColumn, params string[] IndexColumns)
        {
            DataSet ds = SelectOleDbSrvRows(query);
            DataTable dt = ds.Tables[0];
            p.SetBinding(dt.AsEnumerable(), valueColumn, IndexColumns);
        }


        /// <summary>
        /// Load parameters into OML model
        /// </summary>
        private void loadParams()
        {
            foreach (Parameter p in context.CurrentModel.Parameters)
            {
                switch (p.Name)
                {
                    case "AdjRet":
                        setBinding(p, "select [year],instrument,adjreturn from adjreturns",
                            "adjreturn", "year", "instrument");
                        break;
                    case "Mean":
                        setBinding(p, "select instrument,mean from mean",
                            "mean", "instrument");
                        break;
                }

            }
        }

        /// <summary>
        /// get some counts from the database
        /// </summary>
        private void loadScalars()
        {
            OleDbCommand dbcommand = new OleDbCommand("select [value] from counts where [key]='years'", conn);
            numYears = (int)dbcommand.ExecuteScalar();
            dbcommand = new OleDbCommand("select [value] from counts where [key]='instruments'", conn);
            numInstruments = (int)dbcommand.ExecuteScalar();
        }

        /// <summary>
        /// get number of time periods
        /// </summary>
        /// <returns></returns>
        public int getNumt()
        {
            return numYears;
        }

        /// <summary>
        /// get number of instruments
        /// </summary>
        /// <returns></returns>
        public int getNumInstruments()
        {
            return numInstruments;
        }

        /// <summary>
        /// get instruments as 
        /// </summary>
        /// <returns></returns>
        private string[] getInstrumentsFromDB()
        {
            List<string> s = new List<string>();

            OleDbCommand dbcommand = new OleDbCommand("select instrument from instruments", conn);
            OleDbDataReader reader = dbcommand.ExecuteReader();
            while (reader.Read())
            {
                s.Add(reader[0].ToString());
            }
            reader.Close();
            return s.ToArray();
        }

        /// <summary>
        /// get some data from DB
        /// </summary>
        private void initDataFromDB()
        {
            loadScalars();
            instruments = getInstrumentsFromDB();
        }

        /// <summary>
        /// Solve the QP problem
        /// </summary>
        /// <param name="plambda">Value of lambda</param>
        /// <returns>true on success</returns>
        public Boolean SolveQP(double plambda)
        {
            try
            {
                lambda = plambda;
                context.ClearModel();
                if (conn == null)
                {
                    conn = new OleDbConnection(connection);
                    conn.Open();
                    initDataFromDB();
                }
                string s = String.Format(strModelQP, lambda, numYears);
                context.LoadModel(FileFormat.OML, new StringReader(s));
                loadParams();
                solution = context.Solve();
                return true;
            }
            catch(Exception e)
            {
                MessageBox.Show(string.Format("Exception: {0}", e.Message));
                return false;
            }
        }

        /// <summary>
        /// Solve the QP problem
        /// </summary>
        /// <param name="plambda">Value of lambda</param>
        /// <returns>true on success</returns>
        public Boolean SolveLP(double plambda)
        {
            try
            {
                lambda = plambda;
                context.ClearModel();
                if (conn == null)
                {
                    conn = new OleDbConnection(connection);
                    conn.Open();
                    initDataFromDB();
                }
                string s = String.Format(strModelLP, lambda, numYears);
                context.LoadModel(FileFormat.OML, new StringReader(s));
                loadParams();
                solution = context.Solve();
                return true;
            }
            catch (Exception e)
            {
                MessageBox.Show(string.Format("Exception: {0}", e.Message));
                return false;
            }
        }


        /// <summary>
        /// get solution vector
        /// </summary>
        /// <param name="varx"></param>
        public void getSolution(ref double[] varx, ref double returnx, ref double riskx)
        {
            try
            {
                context.PropagateDecisions();
                int n = instruments.Count();
                varx = new double[n];
                Decision x = context.CurrentModel.Decisions.First(d => d.Name == "x");
                for (int i=0; i < n; ++i)
                    varx[i] = x.GetDouble(instruments[i]);
                Decision xreturn = context.CurrentModel.Decisions.First(d => d.Name == "Return");
                returnx = xreturn.ToDouble();
                double obj = solution.Goals.ElementAt(0).ToDouble();
                riskx = (obj + returnx)*numYears/lambda;
                
            }
            catch (Exception e)
            {
                MessageBox.Show(string.Format("Exception: {0}", e.Message));
            }

        }
       
        /// <summary>
        /// get name
        /// </summary>
        /// <param name="i"></param>
        /// <returns></returns>
        public String xname(int i)
        {
            return instruments[i];
        }
         
         
        /// <summary>
        /// Returns the model version 
        /// </summary>
        /// <returns>Version string</returns>
        public String version()
        {
            return versionString;
        }


    }
}
